MODULE instrument_constants
  ! SIZE: buffer size = SIZE*work_index
  integer(KIND=4), parameter :: SIZE      = 1024*1024
  ! NUM_WORKS: number of do_works done
  integer(KIND=4), parameter :: NUM_WORKS = 5
  ! some time spend doing something in do_work
  integer(KIND=4), parameter :: LOOPS     = 1
end MODULE instrument_constants
! -----------------------------------------------------------------
! program that loops through "work" routines (allocates space) then
! frees the memory.
! -----------------------------------------------------------------
program instrument_f
  use instrument_constants
  IMPLICIT NONE
  interface
     subroutine do_work( buf_size_factor, buf )
       integer(KIND=4), intent(in) :: buf_size_factor
       character, dimension(:), pointer :: buf
     end subroutine do_work
  end interface
  integer(KIND=4) :: i           ! loop var
  integer(KIND=4) :: ierr        ! error return value
  integer(KIND=4) :: pinfo_print ! whether or not to print stuff
  real(KIND=8) :: p_wall(2)      ! secs wall clock time (cur, delta)
  real(KIND=8) :: p_time(2)      ! secs proc time (cur, delta)
  real(KIND=8) :: p_size(2)      ! byte proc size (cur, delta)
  real(KIND=8) :: size_tot       ! total size allocated
  type :: char_pointer
     character, dimension(:), pointer :: p
  end type char_pointer
  type (char_pointer), dimension(num_works) :: &
                  bufs           ! array of buffs allocated by do_work
  ! print every time USAGE_WRAPPER is called
  pinfo_print = 1
  ! print header
  write( *, * ) 'Fortran'
  write( *, 100 ) &
          "p_wall now", "delta", &
          "p_time now", "delta", &
          "p_size now", "delta", &
          "expected"
  ! loop of get info , do_work, get info, print changes
  size_tot = 0
  do i = 1, NUM_WORKS
     size_tot = size_tot + i*SIZE
     call USAGE_WRAPPER( pinfo_print, p_wall, p_time, p_size, ierr )
     call do_work( i, bufs(i)%p )
     call USAGE_WRAPPER( pinfo_print, p_wall, p_time, p_size, ierr )
     write( *, 110 ) &
          p_wall(1), p_wall(2), &
          p_time(1), p_time(2), &
          p_size(1), p_size(2), &
          i*SIZE*1.0 
  enddo
  print *, 'Free Memory'
  ! free buffers
  call USAGE_WRAPPER( pinfo_print, p_wall, p_time, p_size, ierr )
  do i = 1, NUM_WORKS
     deallocate( bufs(i)%p )
  enddo
  call USAGE_WRAPPER( pinfo_print, p_wall, p_time, p_size, ierr )
  write( *, 110 ) &
       p_wall(1), p_wall(2), &
       p_time(1), p_time(2), &
       p_size(1), p_size(2), &
       size_tot*(-1.0)
  print *, ' '
  ! FORMATS
100 FORMAT( A10,    ' [', A10,    '] ', A10,    ' [', A10,    '] ', A10, &
         ' [', A10,    '] : [', A10,    ']' )
110 FORMAT( ES10.2, ' [', ES10.2, '] ', ES10.2, ' [', ES10.2, '] ', ES10.2, &
         ' [', ES10.2, '] : [', ES10.2, ']' )
end program instrument_f
! allocate space and spend time doing something
subroutine do_work( &
     buf_size_factor, &
     buf &
  )
  use instrument_constants
  IMPLICIT NONE
  integer(KIND=4), intent(in) :: buf_size_factor
  character, dimension(:), pointer :: buf
  integer(KIND=4) :: i ! loop var
  integer(KIND=4) :: j ! loop var
  ! allocate space
  allocate( buf( 1:SIZE*buf_size_factor ) );
  ! spend time doing something
  do j = 1, LOOPS
     do i = 1, SIZE*buf_size_factor
        buf(i) = 'a'
     enddo
  enddo

end subroutine do_work
