#define _PROCMON_SYNC
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <signal.h>
#include <unistd.h>
#include <assert.h>
#include "procmon_info.h"
/* ------------- */
/* include files */
/* ------------- */
#if defined( _PROCMON_UPS )
#include "ups.h"
#else
#if defined( _PROCMON_SERIAL )
#else
#include "mpi.h"
#endif
#endif
/* ITERATIONS = number of outer loops */
#define ITERATIONS 500
#define PAGES_ALLOC 1
/* NUM_SAVES = number of buffers obtained before freeing */
#define NUM_SAVES 10000/PAGES_ALLOC
#define WORK 2
/* ------------------------------------------------------------------ */
/* each process loops through creating buffers and pe 0 does not free */
/* ------------------------------------------------------------------ */
int main(int argc, char **argv)
{
  int
    i, /* loop var */
    j, /* loop var */
    k, /* loop var */
    l, /* loop var */
    ierr = 0, /* error return value */
    ierr_all, /* error return value summed*/
    penum=0; /* process number */
  char
    *buf[NUM_SAVES]; /* leaky char array */
  char
    tag[64];
  long long
    allocsize, /* size to alloc each time */
    leaksize, /* total leak size so far */
    pagesize; /* size of page */

  /* init and get process number */
#if defined( _PROCMON_UPS )
  ierr = UPS_AA_Init( argc, argv );
  assert( !ierr );
  PROCMON_rld_reset();
  ierr = UPS_CM_Get_penum( &penum );
  assert( !ierr );
#else
#if defined( _PROCMON_SERIAL )
  penum = 0;
#else
  ierr = MPI_Init( &argc, &argv );
  assert( !ierr );
  PROCMON_rld_reset();
  ierr = MPI_Comm_rank( MPI_COMM_WORLD, &penum );
  assert( !ierr );
#endif
#endif
  sprintf( tag, "pe_%d_%d", penum, getpid() );
  fprintf( stderr, "parallel_leak: %d says hi\n", penum );
  pagesize = getpagesize();
  /* --------------------------------------------------------- */
  /* loop doing allocation and some busy work to trigger timer */
  /* and create mem leak.                                      */
  /* --------------------------------------------------------- */
  leaksize = 0;
  allocsize = PAGES_ALLOC * pagesize;
  PROCMON_info_get_print( "a:main", 0 );
  for( i = 0; i < ITERATIONS; i++ )
    {
      if( i%3 == 0 )
        {
          PROCMON_info_get_print( "b:loop", 0 );
        }
      for( j = 0; j < NUM_SAVES; j++ )
        {
          buf[j] = calloc( allocsize, sizeof( char ) );
          if( buf[j] == NULL )
            {
              fprintf( stderr, "penum [%d] malloc failed\n", penum );
              /* free the buffs for prog to terminate gracefully? */
              for( k = 0; k < j; k++ )
                {
                  free( buf[k] );
                }
              ierr = 1;
              break;
            }
          for( l = 0; l < WORK; l++ )
            {
              for( k = allocsize-1; k >= 0; k-- )
                {
                  buf[j][k]++;
                }
            }
        }
      if( ierr == 0 )
        {
          if( penum != 0 )
            {
              for( j = 0; j < NUM_SAVES; j++ )
                {
                  free( buf[j] );
                }
            }
          else
            {
              leaksize += NUM_SAVES*allocsize;
              fprintf( stderr, "%d: leak %3d/%3d size= %15llu\n",
                       penum, i+1, ITERATIONS, leaksize );
            }
        }
      if( i%3 == 0 )
        {
          PROCMON_info_get_print( "b:loop", 1 );
        }
#if defined( _PROCMON_SYNC )
#if defined( _PROCMON_UPS )
      UPS_CM_Allreduce( &ierr, &ierr_all, 1, UPS_DT_INT,
                        UPS_AA_SUM );
#else
#if defined( _PROCMON_SERIAL )
#else
      MPI_Allreduce( &ierr, &ierr_all, 1, MPI_INT,
                     MPI_SUM, MPI_COMM_WORLD );
      if( ierr_all != 0 )
        {
          break;
        }
#endif
#endif
#endif
    }
  PROCMON_info_get_print( "a:main", 1 );
  /* terminate */
#if defined( _PROCMON_UPS )
  ierr = UPS_AA_Terminate();
  assert( !ierr );
#else
#if defined( _PROCMON_SERIAL )
  penum = 0;
#else
  ierr = MPI_Finalize();
  assert( !ierr );
#endif
#endif
  return( ierr );
}
