/* ------------- */
/* include files */
/* ------------- */
#include "procmon_info.h"
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#if defined( _PROCMON_UPS )
#include "ups.h"
#else
#if defined( _PROCMON_SERIAL )
#else
#include "mpi.h"
#endif
#endif
#define NUM_OPS 1000
int main(int argc, char **argv)
{
  int
    i, /* loop var */
    ierr = 0,
    j, /* loop var */
    numpes,
    penum;
  PROCMON_INFO_struct
    pinfo,
    pinfo_time;
  double
    time,
    time_min,
    time_max,
    time_mean,
    *times;
  
  /* init and get process number */
#if defined( _PROCMON_UPS )
  ierr = UPS_AA_Init( argc, argv );
  assert( !ierr );
  ierr = UPS_CM_Get_penum( &penum );
  assert( !ierr );
  ierr = UPS_CM_Get_numpes( &numpes );
  assert( !ierr );
#else
#if defined( _PROCMON_SERIAL )
  penum = 0;
  numpes = 1;
#else
  ierr = MPI_Init( &argc, &argv );
  assert( !ierr );
  ierr = MPI_Comm_rank( MPI_COMM_WORLD, &penum );
  assert( !ierr );
  ierr = MPI_Comm_size( MPI_COMM_WORLD, &numpes );
  assert( !ierr );
#endif
#endif

  PROCMON_info_init( &pinfo );
  PROCMON_info_init( &pinfo_time );
  times = (double*)malloc( numpes * sizeof( double ) );
  /* --- */
  /* Get */
  /* --- */
  if( penum == 0 )
    {
      printf( "# Dataset: Get\n" );
      printf( "# Coord X: numpes\n" );
      printf( "# Coord Y: mean\n" );
      printf( "# NUM_OPS = %d\n", NUM_OPS );
      printf( "# %12s %24s\n", "numpes", "Time/OP" );
      printf( "# %12s %12s %12s %12s\n", "", "Mean", "Min", "Max" );
    }
  for( i = 1; i <= numpes; i*=2 )
    {
      MPI_Barrier( MPI_COMM_WORLD );
      while( PROCMON_info_get( 0, &pinfo_time ) == PROCMON_WARNING ) {}
      time = -pinfo_time.wall;
      for( j = 0; j < NUM_OPS; j++ )
        {
          if( penum < i )
            {
              PROCMON_info_get( 0, &pinfo );
            }
        }
      while( PROCMON_info_get( 0, &pinfo_time ) == PROCMON_WARNING ) {}
      time += pinfo_time.wall;
      time /= NUM_OPS;
      MPI_Gather( &time, 1, MPI_DOUBLE, times, 1, MPI_DOUBLE, 0,
                  MPI_COMM_WORLD );
      if( penum == 0 )
        {
          time_min  = times[0];
          time_max  = times[0];
          time_mean = times[0];
          for( j = 1; j < i; j++ )
            {
              if( times[j] < time_min )
                {
                  time_min = times[j];
                }
              if( times[j] > time_max )
                {
                  time_max = times[j];
                }
              time_mean += times[j];
            }
          time_mean /= j;
          printf( "  %12d %12.5e %12.5e %12.5e\n",
                  i, time_mean, time_min, time_max );
        }
    }
  /* ---- */
  /* GetP */
  /* ---- */
  if( penum == 0 )
    {
      printf( "# Dataset: GetP\n" );
      printf( "# Coord X: numpes\n" );
      printf( "# Coord Y: mean\n" );
      printf( "# NUM_OPS = %d\n", NUM_OPS );
      printf( "# %12s %24s\n", "numpes", "Time/OP" );
      printf( "# %12s %12s %12s %12s\n", "", "Mean", "Min", "Max" );
    }
  for( i = 1; i <= numpes; i*=2 )
    {
      MPI_Barrier( MPI_COMM_WORLD );
      while( PROCMON_info_get( 0, &pinfo_time ) == PROCMON_WARNING ) {}
      time = -pinfo_time.wall;
      for( j = 0; j < NUM_OPS; j++ )
        {
          if( penum < i )
            {
              PROCMON_info_get_print( "a", 0 );
            }
        }
      while( PROCMON_info_get( 0, &pinfo_time ) == PROCMON_WARNING ) {}
      time += pinfo_time.wall;
      time /= NUM_OPS;
      MPI_Gather( &time, 1, MPI_DOUBLE, times, 1, MPI_DOUBLE, 0,
                  MPI_COMM_WORLD );
      if( penum == 0 )
        {
          time_min  = times[0];
          time_max  = times[0];
          time_mean = times[0];
          for( j = 1; j < i; j++ )
            {
              if( times[j] < time_min )
                {
                  time_min = times[j];
                }
              if( times[j] > time_max )
                {
                  time_max = times[j];
                }
              time_mean += times[j];
            }
          time_mean /= j;
          printf( "  %12d %12.5e %12.5e %12.5e\n",
                  i, time_mean, time_min, time_max );
        }
    }
  free( times );
  /* terminate */
#if defined( _PROCMON_UPS )
  ierr = UPS_AA_Terminate();
  assert( !ierr );
#else
#if defined( _PROCMON_SERIAL )
  penum = 0;
#else
  ierr = MPI_Finalize();
  assert( !ierr );
#endif
#endif
  return( ierr );
}
