eval 'exec perl -w -S $0 ${1+"$@"}'
  if 0;

if ( $#ARGV != 0 || $ARGV[0] =~ /^--?help/)
  {
    print <<'EOF';
#............................................................................
#...Name
#...====
#... fit_plot.pl
#...
#...Usage
#...=====
#... fit_plot.pl
#...   <file>
#...
#... Options
#... -------
#...  <file>
#...    The text data file containing the dataset(s) to fit/plot.
#...    Format:
#...        # Dataset: <Name>
#...        [# flag: newpage] : start this plot on a new page
#...        # Coord X: <coord X name>
#...        # Coord Y: <coord Y name>
#...        <coord 1 data>  <coord 2 data>
#...        <coord 1 data>  <coord 2 data>
#...        <...repeat data...>
#...        <...repeat for other datasets...>
#...    Note: other lines starting with "#" will be ignored.
#...
#...    Example:
#...      # Dataset: foo
#...      # Coord X: time
#...      # Coord Y: distance
#...      # 2 cols of data
#...         1    2
#...         3    4
#...      # Dataset: bar
#...      # flag: newpage
#...      # Coord X: dollars
#...      # Coord Y: hours
#...      # some more data
#...         5    6
#...         7    8
#...      # Dataset: hmm
#...      # Coord X: heat
#...      # Coord Y: kitchen
#...      # some more data
#...         9    10
#...         11   12
#...         13   14
#...
#...Purpose
#...=======
#... Fit curve to each dataset (return results) and make plots.
#... Currently, only linear fits using linear least squares.
#...
#... This script uses gnuplot and ps2pdf (they must be in the PATH).
#...
#...Examples
#...========
#... 1) Fit/plot data:
#...      fit_plot.pl data.txt
#............................................................................
EOF
  exit;
  }
#............................................................................
#...Program Flow
#...============
#...1) read file
#...2) compute fit
#...3) print fit
#...4) plot data/fit
#............................................................................
#.............
#...Globals...
#.............
$GMIN        = 1e99;
$GMAX        = -1e99;
$GNAME       = "name";
$GVAL        = "val";
$GPARAM      = "param";
$GCOORD_X    = "X";
$GCOORD_Y    = "Y";
$GFIT_LINEAR = "linear";
#.............
#...my vars...
#.............
my(
   %data, # the data from input file:
         # $data{$ds}{$GCOORD_X}{name} = name of coordinate
         # $data{$ds}{$GCOORD_X}{vals} = array of values
   $ierr, # int error
   %fit, # fit: Y = aX + b;
         # $fit{$GFIT_LINEAR}{a}
   # $fit{$GFIT_LINEAR}{b}
  );
#...............
#...read file...
#...............
&read_file( $ARGV[0], \%data );
#.................
#...compute fit...
#.................
&fit_data( \%data, \%fit );
#...............
#...print fit...
#...............
&print_fit( \%data, \%fit );
#...................
#...plot data/fit...
#...................
&plot_data_fit( \%data, \%fit );
exit( 0 );
#............................................................................
#............................ Subroutines ...................................
#............................................................................
sub numerically{$a <=> $b;}
#............................................................................
#...Name
#...====
#... plot_data_fit
#...
#...Purpose
#...=======
#... Plots the data and the fit values
#...
#...Arguments
#...=========
#... data_ref          Intent: in
#...                   Perl type: reference to hash
#...                   reference to %data
#...
#... fit_ref           Intent: out
#...                   Perl type: reference to hash
#...                   reference %fit hash
#...
#...Program Flow
#...============
#... 1) print_data_file
#... 2) create_plot
#............................................................................
sub plot_data_fit
  {
    my(
       $data_ref,
       $fit_ref,
      ) = @_;
    my(
       $ds_name, # dataset name
       %gnuplot_info, # info that will be used to plot data
      );
    &print_data_file( $data_ref, $fit_ref, \%gnuplot_info );
    create_plot( \%gnuplot_info );
  }
#............................................................................
#...Name
#...====
#... create_plot
#...
#...Purpose
#...=======
#... Given the gnuplot_info_ref, create the plots
#...
#...Arguments
#...=========
#... gnuplot_info_ref  Intent: in
#...                   Perl type: reference to hash
#...                   reference to %gnuplot_info
#...
#...Program Flow
#...============
#... 1) find execs
#... 2) plot each dataset
#... 3) print cmd file
#... 4) run gnuplot
#............................................................................
sub create_plot
  {
    my(
       $gnuplot_info_ref,
      ) = @_;
    my(
       $arch, # architecture
       $ds_name, # dataset name
       $gnuplot, # where gnuplot is
       $gnuplot_cmd, # the gnuplot command to print to file
       $gnuplot_origin, # origin
       $gnuplot_size, # size
       $i, # loop var
       $j, # loop var
       $key, # key value
       $ierr, # error return value
       $num_plots_total, # number of plots total to be plotted
       $num_plots_total_done, # number of plots done so far total
       $num_plots_page, # number of plots to be printed this page
       $num_plots_page_done, # number of plots done for this page
       $output_file_cmd, # command file
       $output_file_ps, # output ps file
       $output_file_pdf, # output pdf file
       $ps2pdf, # where ps2pdf is
      );
    $output_file_ps  = "fit_plot.ps";
    $output_file_pdf = "fit_plot.pdf";
    $output_file_cmd = "fit_plot.cmd";
    #................
    #...find execs...
    #................
    $arch = `uname -a 2>&1`;
    if( $arch =~ /alpha/i )
      {
        $ENV{PATH} = "/usr/local/gnu/bin:$ENV{PATH}";
        $ENV{PATH} = "/usr/projects/ups/ghostscript/alpha/bin:$ENV{PATH}";
      }
    if( $arch =~ /irix/i )
      {
        $ENV{PATH} = "/usr/local/gnu/bin:$ENV{PATH}";
        $ENV{PATH} = "/usr/projects/ups/ghostscript/irix/bin:$ENV{PATH}";
      }
    if( $arch =~ /linux/i )
      {
        $ENV{PATH} = "/usr/projects/ups/ghostscript/linux/bin:$ENV{PATH}";
      }
    $gnuplot = &which_exec( "gnuplot" );
    if( $gnuplot !~ /\S/ )
      {
        return;
      }
    $ps2pdf = &which_exec( "ps2pdf" );
    #........................
    #...create gnuplot_cmd...
    #........................
    $gnuplot_cmd = "";
    $gnuplot_cmd .= "# \n";
    $gnuplot_cmd .= "# This file is to be processed by gnuplot.\n";
    $gnuplot_cmd .= "# The data file(s) from fit_plot.pl are read and\n";
    $gnuplot_cmd .= "# $output_file_ps is created.\n";
    $gnuplot_cmd .= "# \n";
    $gnuplot_cmd .= "set output '$output_file_ps'\n";
    $gnuplot_cmd .= "set terminal postscript portrait color 'Times-Roman' 10\n";
    $num_plots_total = $#{$$gnuplot_info_ref{ds_order}} + 1;
    $num_plots_total_done = 0;
    $num_plots_page_done = 0;
    $num_plots_page = 0;
    #.......................
    #...plot each dataset...
    #.......................
    for( $i = 0; $i < $num_plots_total; $i++ )
      {
        $ds_name = $$gnuplot_info_ref{ds_order}[$i];
        $gnuplot_cmd .= "#\n";
        $gnuplot_cmd .= "# Plotting [$$gnuplot_info_ref{data}{$ds_name}{title}]\n";
        if( $num_plots_page_done == $num_plots_page )
          {
            $num_plots_page = 1;
            for( $j = $i+1; $j < $num_plots_total; $j++ )
              {
                $key = $$gnuplot_info_ref{ds_order}[$j];
                if( defined($$gnuplot_info_ref{data}{$key}{flag}{newpage}) )
                  {
                    last;
                  }
                $num_plots_page++;
              }
            $num_plots_page_done = 0;
            $gnuplot_size = 1.0/$num_plots_page;
            $gnuplot_origin = 1 - $gnuplot_size;
            $gnuplot_cmd .= "# New Page\n";
            $gnuplot_cmd .= "set multiplot\n";
          }
        else
          {
            $gnuplot_origin -= $gnuplot_size;
          }
        $gnuplot_cmd .= sprintf( "set size 1, %.4f\n",
                                 abs($gnuplot_size) );
        $gnuplot_cmd .= sprintf( "set origin 0, %.4f\n",
                                 abs($gnuplot_origin) );
        $gnuplot_cmd .= "set title '$$gnuplot_info_ref{data}{$ds_name}{title}'\n";
        $gnuplot_cmd .= "set xlabel '$$gnuplot_info_ref{data}{$ds_name}{xlabel}'\n";
        $gnuplot_cmd .= "set ylabel '$$gnuplot_info_ref{data}{$ds_name}{ylabel}'\n";
        $gnuplot_cmd .= "plot ";
        $gnuplot_cmd .= "'$$gnuplot_info_ref{data}{$ds_name}{filename}' ";
        $gnuplot_cmd .= "index $$gnuplot_info_ref{data}{$ds_name}{data}{orig}{index} ";
        $gnuplot_cmd .= "using $$gnuplot_info_ref{data}{$ds_name}{data}{orig}{using} ";
        $gnuplot_cmd .= "title '$$gnuplot_info_ref{data}{$ds_name}{data}{orig}{title}' ";
        $gnuplot_cmd .= "with points pt 3 ps 1, ";
        $gnuplot_cmd .= "'$$gnuplot_info_ref{data}{$ds_name}{filename}' ";
        $gnuplot_cmd .= "index $$gnuplot_info_ref{data}{$ds_name}{data}{fit}{$GFIT_LINEAR}{index} ";
        $gnuplot_cmd .= "using $$gnuplot_info_ref{data}{$ds_name}{data}{fit}{$GFIT_LINEAR}{using} ";
        $gnuplot_cmd .= "title '$$gnuplot_info_ref{data}{$ds_name}{data}{fit}{$GFIT_LINEAR}{title}' ";
        $gnuplot_cmd .= "with lines lw 2";
        $gnuplot_cmd .= "\n";
        $num_plots_page_done++;
        $num_plots_total_done++;
      }
    #....................
    #...print cmd file...
    #....................
    printf " Creating [$output_file_cmd]\n";
    if( !open( FILE, ">$output_file_cmd" ) )
      {
        $ierr = 1;
        &print_error( "Cannot open output command file [$output_file_cmd].",
                      $ierr );
        exit( $ierr );
      }
    print FILE $gnuplot_cmd;
    close( FILE );
    printf " Creating [$output_file_ps]\n";
    $output = `$gnuplot $output_file_cmd > /dev/null 2>&1`;
    print $output;
    if( $ps2pdf ne "" && -T "$output_file_ps" )
      {
        printf " Creating [$output_file_pdf]\n";
        $output = `$ps2pdf $output_file_ps $output_file_pdf 2>&1`;
        print $output;
      }
  }
#............................................................................
#...Name
#...====
#... print_data_file
#...
#...Purpose
#...=======
#... Prints the file gnuplot will use and fills gnuplot_info_ref
#...
#...Arguments
#...=========
#... data_ref          Intent: in
#...                   Perl type: reference to hash
#...                   reference to %data
#...
#... fit_ref           Intent: in
#...                   Perl type: reference to hash
#...                   reference to %fit
#...
#... gnuplot_info_ref  Intent: inout
#...                   Perl type: reference to hash
#...                   reference to %gnuplot_info
#...
#...Program Flow
#...============
#... 1) foreach dataset
#... 1.1) print original data (and assign gnuplot_info_ref)
#... 1.2) print fit      data (and assign gnuplot_info_ref)
#............................................................................
sub print_data_file
  {
    my(
       $data_ref,
       $fit_ref,
       $gnuplot_info_ref
      ) = @_;
    my(
       $coord, # coordinate label
       $coord_last, # last coordinate label
       $ds_name, # dataset name
       $i, # loop var
       $ierr, # error return value
       $index, # gnuplot value
       $key, # a key in a hash
       $num_points, # number of points
       $outfile_name, # output file
       $title, # a title
      );
    $outfile_name = "fit_plot.dat";
    if( !open( FILE, ">$outfile_name" ) )
      {
        $ierr = 1;
        &print_error( "Cannot open output data file [$outfile_name].",
                      $ierr );
        exit( $ierr );
      }
    $index = 0;
    foreach $ds_name ( @{$$data_ref{ds_order}} )
      {
        push( @{$$gnuplot_info_ref{ds_order}}, $ds_name );
        foreach $key ( keys %{$$data_ref{data}{$ds_name}{flag}} )
          {
            $$gnuplot_info_ref{data}{$ds_name}{flag}{$key} = "";
          }
        $$gnuplot_info_ref{data}{$ds_name}{filename} = $outfile_name;
        #.........................................
        #...gnuplot_info: title, xlabel, ylabel...
        #.........................................
        $title = "$ds_name";
        $$gnuplot_info_ref{data}{$ds_name}{title} = $title;
        $i = 0;
        foreach $coord ( sort keys %{$$data_ref{data}{$ds_name}{coord}} )
          {
            if( $i == 0 )
              {
                $key = "xlabel";
              }
            else
              {
                $key = "ylabel";
              }
            $$gnuplot_info_ref{data}{$ds_name}{$key} =
              $$data_ref{data}{$ds_name}{coord}{$coord}{name};
            $i++;
          }
        #.........................
        #...print original data...
        #.........................
        $title = "Original Data";
        $$gnuplot_info_ref{data}{$ds_name}{data}{orig}{title} = $title;
        $$gnuplot_info_ref{data}{$ds_name}{data}{orig}{index} = $index;
        $$gnuplot_info_ref{data}{$ds_name}{data}{orig}{using} = "1:2";
        $index++;
        printf FILE "# Dataset: %s\n", $ds_name;
        printf FILE "# Original Data\n";
        $coord_last = "";
        foreach $coord ( sort keys %{$$data_ref{data}{$ds_name}{coord}} )
          {
            printf FILE "# Coord %s: %s\n",
              $coord, $$data_ref{data}{$ds_name}{coord}{$coord}{name};
            $coord_last = $coord;
          }
        print FILE "#";
        foreach $coord ( sort keys %{$$data_ref{data}{$ds_name}{coord}} )
          {
            printf FILE " %17s", $coord;
          }
        print FILE "\n";
        $num_points = $#{$$data_ref{data}{$ds_name}{coord}{$coord_last}{val}};
        for( $i = 0; $i <= $num_points; $i++ )
          {
            print FILE " ";
            foreach $coord ( sort keys %{$$data_ref{data}{$ds_name}{coord}} )
              {
                printf FILE " %17.6e",
                  $$data_ref{data}{$ds_name}{coord}{$coord}{val}[$i];
              }
            print FILE "\n";
          }
        #....................
        #...print fit data...
        #....................
        $title = sprintf( "Linear: Y = %17.6e * X + %17.6e",
                          $$fit_ref{$ds_name}{$GFIT_LINEAR}{param}{a},
                          $$fit_ref{$ds_name}{$GFIT_LINEAR}{param}{b} );
        $$gnuplot_info_ref{data}{$ds_name}{data}{fit}{$GFIT_LINEAR}{title} =
          $title;
        $$gnuplot_info_ref{data}{$ds_name}{data}{fit}{$GFIT_LINEAR}{index} =
          $index;
        $$gnuplot_info_ref{data}{$ds_name}{data}{fit}{$GFIT_LINEAR}{using} =
          "1:2";
        $index++;
        print FILE "\n\n";
        printf FILE "# $title\n";
        $num_points = $#{$$fit_ref{$ds_name}{$GFIT_LINEAR}{val}{$coord_last}};
        for( $i = 0; $i <= $num_points; $i++ )
          {
            print FILE " ";
            foreach $coord ( sort keys %{$$fit_ref{$ds_name}{$GFIT_LINEAR}{val}} )
              {
                printf FILE " %17.6e",
                  $$fit_ref{$ds_name}{$GFIT_LINEAR}{val}{$coord}[$i];
              }
            print FILE "\n";
          }
        print FILE "\n\n";
      }
    close( FILE );
  }
#............................................................................
#...Name
#...====
#... print_fit
#...
#...Purpose
#...=======
#... Prints the results of the fit to screen
#...
#...Arguments
#...=========
#... data_ref          Intent: in
#...                   Perl type: reference to hash
#...                   reference to %data
#...
#... fit_ref           Intent: in
#...                   Perl type: reference to hash
#...                   reference to %fit
#...
#...Program Flow
#...============
#... 1) foreach dataset
#... 1.1) print fid data
#............................................................................
sub print_fit
  {
    my(
       $data_ref,
       $fit_ref,
      ) = @_;
    my(
       $ds_name, # dataset name
      );
    foreach $ds_name ( @{$$data_ref{ds_order}} )
      {
        print "Dataset: $ds_name\n";
        print "  X = $$data_ref{data}{$ds_name}{coord}{$GCOORD_X}{$GNAME}\n";
        print "  Y = $$data_ref{data}{$ds_name}{coord}{$GCOORD_Y}{$GNAME}\n";
        printf( "  Y = %14.7e * X + %14.7e\n",
                $$fit_ref{$ds_name}{$GFIT_LINEAR}{$GPARAM}{a},
                $$fit_ref{$ds_name}{$GFIT_LINEAR}{$GPARAM}{b} );
      }
  }
#............................................................................
#...Name
#...====
#... fit_data
#...
#...Purpose
#...=======
#... Fit the data
#...
#...Arguments
#...=========
#... data_ref          Intent: in
#...                   Perl type: reference to hash
#...                   reference to %data
#...
#... fit_ref           Intent: inout
#...                   Perl type: reference to hash
#...                   reference to %fit
#...
#...Program Flow
#...============
#... 1) foreach dataset
#... 1.1) do linear least squares fit
#............................................................................
sub fit_data
  {
    my(
       $data_ref,
       $fit_ref,
      ) = @_;
    my(
       $ds_name,    # dataset name
       $a,          # slope
       $b,          # y-intcp
       $den,        # denominator
       $i,          # loop var
       $min_x,      # min x value
       $max_x,      # max x value
       $sx,         # sum x
       $sxx,        # sum x squared
       $sxy,        # sum xy
       $sy,         # sum y
       $syy,        # sum y squared
       $x,          # x value
       @x_vals,     # array of x values
       $y,          # y value
       $num_points, # number of points
      );
    #..........................
    #...process each dataset...
    #..........................
    foreach $ds_name ( keys %{$$data_ref{data}} )
      {
        $num_points = $#{$$data_ref{data}{$ds_name}{coord}{$GCOORD_X}{$GVAL}} + 1;
        if( $num_points < 2 )
          {
            $ierr = 0;
            &print_error( "Must have at least 2 points [$num_points].",
                          "Dataset: [$ds_name].",
                          $ierr );
            exit( $ierr );
          }
        #..................
        #...compute sums...
        #..................
        $sx  = 0;
        $sxx = 0;
        $sxy = 0;
        $sy  = 0;
        $syy = 0;
        $min_x = $GMIN;
        $max_x = $GMAX;
        for( $i = 0; $i < $num_points; $i++ )
          {
            $x = $$data_ref{data}{$ds_name}{coord}{$GCOORD_X}{$GVAL}[$i];
            $y = $$data_ref{data}{$ds_name}{coord}{$GCOORD_Y}{$GVAL}[$i];
            $sx  += $x;
            $sxx += $x**2;
            $sxy += $x * $y;
            $sy  += $y;
            $syy += $y ** 2;
            if( $x < $min_x )
              {
                $min_x = $x;
              }
            if( $x > $max_x )
              {
                $max_x = $x;
              }
          }
        $den = $num_points * $sxx - $sx**2;
        if( $den == 0 )
          {
            $ierr = 1;
            &print_error( "Denominator of LLS 0 - invalid",
                          "Dataset: [$ds_name].",
                          $ierr );
            exit( $ierr );
          }
        #.........................
        #....find a,b: Y=aX + b...
        #.........................
        $a = ($num_points*$sxy - $sx*$sy) / $den;
        $b = ($sxx*$sy - $sxy*$sx) / $den;
        $$fit_ref{$ds_name}{$GFIT_LINEAR}{$GPARAM}{a} = $a;
        $$fit_ref{$ds_name}{$GFIT_LINEAR}{$GPARAM}{b} = $b;
        @x_vals = sort numerically (0, $min_x, $max_x);
        foreach $x ( @x_vals )
          {
            $y = $a*$x + $b;
            push( @{$$fit_ref{$ds_name}{$GFIT_LINEAR}{$GVAL}{$GCOORD_X}}, $x );
            push( @{$$fit_ref{$ds_name}{$GFIT_LINEAR}{$GVAL}{$GCOORD_Y}}, $y );
          }
      }
    #................................
    #...DONE: process each dataset...
    #................................
  }
#............................................................................
#...Name
#...====
#... read_file
#...
#...Purpose
#...=======
#... Read the file into data_ref
#...
#...Arguments
#...=========
#... filename          Intent: in
#...                   Perl type: scalar
#...                   the file to read in
#...
#... data_ref          Intent: inout
#...                   Perl type: reference to hash
#...                   reference to %data
#...
#...Program Flow
#...============
#... 1) foreach line
#... 1.1) process dataset, coord, comment, or values
#............................................................................
sub read_file
  {
    my(
       $filename,
       $data_ref,
      ) = @_;
    my(
       $ds_name, # dataset name
       $line, # line of file
       $line_num, # line number of file
       $val, # single @vals
       @vals, # split of data line
      );
    if( ! open( FILE, $filename ) )
      {
        $ierr = 1;
        &print_error( "Cannot open datafile [$filename]",
                      $ierr );
        exit( $ierr );
      }
    $ds_name = "UNKNOWN";
    $line_num = 0;
    while( $line=<FILE> )
      {
        $line_num++;
        #.............
        #...dataset...
        #.............
        if( $line =~
            /
            ^\s*\#\s*     # starts with pound
            Dataset\:\s*  # keword
            (\S.*?)       # value
            \s+$          # whitespace return
            /x )
          {
            $ds_name = $1;
            push( @{$$data_ref{ds_order}}, $ds_name );
          }
        #...........
        #...coord...
        #...........
        elsif( $line =~
               /
               ^\s*\#\s*        # starts with pound
               Coord\s+(\S):\s* # Coord
               (\S.*?)          # value
               \s*$             # whitespace eol
               /x )
          {
            $$data_ref{data}{$ds_name}{coord}{$1}{$GNAME} = $2;
          }
        #.............
        #...newpage...
        #.............
        elsif( $line =~
               /
               ^\s*\#\s*        # starts with pound
               flag:\s*         # flag option
               (\S.*?)          # value
               \s*$             # whitespace eol
               /x )
          {
            $$data_ref{data}{$ds_name}{flag}{$1} = "";
          }
        #.............
        #...comment...
        #.............
        elsif( $line =~ /^\s*\#/ ||
               $line !~ /\S/ )
          {
            #...skip...
          }
        #...............
        #...data line...
        #...............
        else
          {
            $line =~ s/^\s*(\S.*?)\s*$/$1/;
            @vals = split( /\s+/, $line );
            if( $#vals < 1 )
              {
                $ierr = 1;
                &print_error( "Must have at least 2 values per line.",
                              "Other values will be ignored.",
                              "File: [$filename:$line_num].",
                              "Line: [$line]",
                              $ierr );
                exit( $ierr );
              }
            push( @{$$data_ref{data}{$ds_name}{coord}{$GCOORD_X}{$GVAL}}, $vals[0] );
            push( @{$$data_ref{data}{$ds_name}{coord}{$GCOORD_Y}{$GVAL}}, $vals[1] );
          }
      }
    close( FILE );
  }
#............................................................................
#...Name
#...====
#... print_error
#...
#...Purpose
#...=======
#... Print a standard error message.
#...
#...Arguments
#...=========
#... error_lines  Intent: in
#...              Perl type: array
#...              0: cause of error (file_name:line_number)
#...              1: explanation/fix (if there is one)
#...              2: error value
#...
#...Program Flow
#...============
#... 1) see if warning or error (last argument is 0 or not)
#... 2) Line up error lines by column
#... 3) find out who was calling this routine and the line number
#... 4) print out info
#............................................................................
sub print_error
  {
    my(
       @error_lines # what emitted the error
      ) = @_;
    my(
       $error_level,  # what is printed (eg WARNING or ERROR)
       $error_line,   # each line of input argument
       @routine_info, # var from caller
       $routine_name, # routine name
       $spaces        # spaces for lining up columns
      );
    #......................................................
    #...assign WARNING or ERROR depending on error value...
    #......................................................
    if ( "$error_lines[$#error_lines]" eq "0" )
      {
        $error_level = "**WARNING**";
        $spaces      = "           ";
      }
    else
      {
        $error_level = "**ERROR**";
        $spaces      = "         ";
      }
    #............................................................
    #...DONE: assign WARNING or ERROR depending on error value...
    #............................................................
    #.......................................
    #...init error and add argument lines...
    #.......................................
    $error_line = shift( @error_lines );
    $error_message = "\n$error_level $error_line\n";
    foreach $error_line ( @error_lines )
      {
        $error_message .= "$spaces $error_line\n";
      }
    #............................................................
    #...get routine name (special if called from main program)...
    #............................................................
    @routine_info = caller(1);
    if ( $#routine_info < 3 )
      {
        $routine_name = $0;
      }
    else
      {
        $routine_name = (caller(1))[3];
      }
    $routine_name = $routine_name.":".(caller(0))[2];
    #...........................
    #...postpend routine name...
    #...........................
    $error_message .= "$spaces $routine_name\n";
    #.................
    #...print error...
    #.................
    print STDERR $error_message;
  }
#............................................................................
#...Name
#...====
#... which_exec
#...
#...Purpose
#...=======
#... Returns value for full path to executable ("" if not found).
#...
#...Arguments
#...=========
#... exec Intent: in
#...      Perl type: string scalar
#...
#...Program Flow
#...============
#... 1) go through path to find exec
#............................................................................
sub which_exec
  {
    my( 
       $exec, # exec to find
      ) = @_;
    my(
       $ierr, # error return value
       $exec_try, # see if exec is here
       $found, # if found
       $path, # current dir in search for execs
       @paths, # list of search paths for execs
       $this_dir, # current directory
      );
    #.................................................
    #...build paths from PATH and current directory...
    #.................................................
    $path = $ENV{PATH};
    @paths = split( /:/, $ENV{PATH} );
    ($this_dir = $0) =~ s/\/[^\/]+$//;
    unshift( @paths, $this_dir );
    #.....................................
    #...loop through paths to find exec...
    #.....................................
    $found = "false";
    foreach $path (@paths)
      {
        $exec_try = "$path/$exec";
        if( -x "$exec_try" )
          {
            $found = "true";
            last;
          }
      }
    #........................................
    #...error if still could not find exec...
    #........................................
    if( $found eq "false" )
      {
        $ierr = 1;
        &print_error(
                     "Necessary executable [$exec] not found in PATH",
                     $ierr
                    );
        $exec_try = "";
      }
    return( $exec_try );
  }
#............................................................................
#...Name
#...====
#... print_perl_obj
#...
#...Purpose
#...=======
#... Print the structure of a perl object...for easier debugging.
#... Mixtures of arrays, hashes, and scalars can be printed.
#...
#... Up to a certain number of hash/array values are printed.
#... Beyond this, values are skipped (unless they are a reference
#... themselves) until the last value of the array/hash.
#...
#...Arguments
#...=========
#... $obj_ref          Intent: in
#...                   Perl type: reference to array
#...                   Reference to object to print (\%, \@, \$)
#...
#... $in_pref          Intent: in
#...                   Perl type: scalar
#...                   The preface string to name this object.
#...                   If passing \%foo, a good pref might be "foo".
#...                   if not defined, a default will be used.
#...
#... $in_max_items     Intent: in
#...                   Perl type: scalar
#...                   Maximum number of items in array/hash to print.
#...                   If negative, print all items.
#...                   If not defined, all items will be printed.
#...
#... $in_file          Intent: in
#...                   Perl type: scalar
#...                   Output file (usually just STDOUT).
#...                   on the same plot.
#...                   If not defined, STDOUT will be used.
#...
#...Program Flow
#...============
#... 1) If hash or array, recursively call this routine.
#... 2) If scalar, print the value.
#............................................................................
sub print_perl_obj
{
  my(
     $obj_ref,
     $in_pref,
     $in_max_items,
     $in_file
    ) = @_;
  my(
     $file, # file to use
     $i, # loop var
     @indices, # indices of object
     $j, # loop var
     $key, # key
     $max_items, # max items to use
     $new_pref, # new preface
     $pref, # preface to use
     $skip, # if skipping
     $skip_count, # how many skipped
     $val, # scalar value
    );
  #..........................
  #...fix non-defined vals...
  #..........................
  if( ! defined( $in_pref ) )
    {
      $pref = "var";
    }
  else
    {
      $pref = $in_pref;
    }
  if( ! defined( $in_max_items ) )
    {
      $max_items = -1;
    }
  else
    {
      $max_items = $in_max_items;
    }
  if( ! defined( $in_file ) )
    {
      $file = STDOUT;
    }
  else
    {
      $file = $in_file;
    }
  #...............................
  #...print depending upon type...
  #...............................
  if( ref( $obj_ref ) eq "ARRAY" )
  {
    if( $#{$obj_ref} < 0 )
      {
        print {$file} "$pref\[\] undefined\n";
      }
    else
      {
        $skip = 0;
        $skip_count = 0;
        for( $i = 0; $i <= $#{$obj_ref}; $i++ )
          {
            #...determine if should skip printing this one...
            if( $i > $max_items-1 && $max_items >= 0 )
              {
                $skip = 1;
              }
            $new_pref = sprintf( "%s[%s]", $pref, $i );
            $val = $$obj_ref[$i];
            if( ref( $val ) )
              {
                #...print number skipped...
                if( $skip_count > 0 )
                  {
                    print {$file} "$pref\[...$skip_count\]\n";
                    $skip_count = 0;
                  }
                &print_perl_obj( $val, $new_pref, $max_items, $file )
              }
            else
              {
                #...print last one no matter what...
                if( $skip && $i < $#{$obj_ref} )
                  {
                    $skip_count++;
                    next;
                  }
                if( $skip_count > 0 )
                  {
                    print {$file} "$pref\[...$skip_count\]\n";
                    $skip_count = 0;
                  }
                &print_perl_obj( \$val, $new_pref, $max_items, $file )
              }
          }
      }
  }
  elsif( ref( $obj_ref ) eq "HASH" )
  {
    @indices = sort keys %{$obj_ref};
    if( $#indices < 0 )
      {
        print {$file} "$pref\{\} undefined\n";
      }
    else
      {
        $skip = 0;
        $skip_count = 0;
        for( $i = 0; $i <= $#indices; $i++ )
          {
            #...determine if should skip printing this one...
            if( $i > $max_items-1 && $max_items >= 0 )
              {
                $skip = 1;
              }
            $new_pref = sprintf( "%s{%s}", $pref, $indices[$i] );
            $val = $$obj_ref{$indices[$i]};
            if( ref( $val ) )
              {
                #...print number skipped...
                if( $skip_count > 0 )
                  {
                    print {$file} "$pref\{...$skip_count\}\n";
                    $skip_count = 0;
                  }
                &print_perl_obj( $val, $new_pref, $max_items, $file )
              }
            else
              {
                if( $skip && $i < $#indices )
                  {
                    $skip_count++;
                    next;
                  }
                if( $skip_count > 0 )
                  {
                    print {$file} "$pref\{...$skip_count\}\n";
                    $skip_count = 0;
                  }
                &print_perl_obj( \$val, $new_pref, $max_items, $file )
              }
          }
      }
  }
  else
  {
     if( defined( $$obj_ref ) )
     {
       print {$file} "$pref = [$$obj_ref]\n";
     }
     else
     {
       print {$file} "$pref undefined\n";
     }
  }
}
