#include "procmon_info.h"
#include <unistd.h>
#include <time.h>
#include <assert.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>
#include <stdlib.h>
#include <ctype.h>
#include <termio.h>
typedef struct PROCMON_INFO_LIST_struct
{
  struct PROCMON_INFO_LIST_struct *next;
  PROCMON_INFO_struct pinfo;
  FILE *outfile; /* output file pointer */
  char outfile_name[PROCMON_MAX_LEN]; /* output file name */
  long long pid; /* process id for this struct */
  int ierr_pget; /* error return flag from PROCMON_info_get */
  int call_get_info; /* if new PROCMON_info_get should be called */
}
PROCMON_INFO_LIST_struct;
int PROCMON_sleep(
                  double time
                  );
int PROCMON_list_create(
                        PROCMON_INFO_SETTINGS_struct *psettings,
                        PROCMON_INFO_LIST_struct **plist_head )
{
  int ierr=0; /* error return variable */
  int ierr_pget; /* error for PROCMON_info_get */
  int match; /* T/F it search process matches */
  DIR *proc_dir; /* opendir */
  struct dirent *proc_dir_dirent; /* readdir */
  char cwd[PROCMON_MAX_LEN]; /* current working dir */
  long long pid; /* process id for search in /proc */
  PROCMON_INFO_struct pinfo; /* temp for search in /proc */
  PROCMON_INFO_LIST_struct *plist; /* pointer to list */
  /* ----------------------------------------- */
  /* if getting info about a list, create list */
  /* ----------------------------------------- */
  if( psettings->pid == 0 )
    {
      getcwd( cwd, PROCMON_MAX_LEN );
      proc_dir = opendir("/proc");
      while( (proc_dir_dirent = readdir(proc_dir)) != NULL )
        {
          if( isdigit( proc_dir_dirent->d_name[0] ) )
            {
              pid = atoi( proc_dir_dirent->d_name );
              /* ---------------------------------------------------- */
              /* On OSF1, /proc/00000 sometimes gets created - ignore */
              /* ---------------------------------------------------- */
              if( pid == 0 )
                {
                  continue;
                }
              PROCMON_info_init( &pinfo );
              ierr_pget = PROCMON_info_get( pid, &pinfo );
              match = 0;
              if( ierr_pget == PROCMON_ERROR )
                {
                  match = 0;
                }
              else if( psettings->all != 0 )
                {
                  match = 1;
                }
              else if( strlen( psettings->command ) == 0 &&
                  psettings->uid >= 0 &&
                  pinfo.p_uid == psettings->uid )
                {
                  match = 1;
                }
              else if( psettings->uid < 0 &&
                  strlen( psettings->command ) > 0 &&
                  strstr( pinfo.p_command, psettings->command ) != NULL )
                {
                  match = 1;
                }
              else if( psettings->uid >= 0 &&
                  pinfo.p_uid == psettings->uid &&
                  strlen( psettings->command ) > 0 &&
                  strstr( pinfo.p_command, psettings->command ) != NULL )
                {
                  match = 1;
                }
              /* ----------- */
              /* valid entry */
              /* ----------- */
              if( match != 0 )
                {
                  plist = calloc( 1, sizeof( PROCMON_INFO_LIST_struct ) );
                  plist->pid = pid;
                  plist->call_get_info = 0;
                  PROCMON_info_copy( &pinfo, &(plist->pinfo) );
                  if( strlen( psettings->outfile_prefix ) > 0 )
                    {
                      sprintf( plist->outfile_name, "%s%llu%s",
                               psettings->outfile_prefix, plist->pid, ".txt" );
                      plist->outfile = fopen( plist->outfile_name, "a" );
                    }
                  else
                    {
                      plist->outfile = NULL;
                    }
                  plist->next = *plist_head;
                  *plist_head = plist;
                }
            }
        }
    }
  /* ----------------------------------------------- */
  /* DONE: if getting info about a list, create list */
  /* ----------------------------------------------- */
  /* ------------------------------------------------------- */
  /* if getting info about a single pid, just create 1 entry */
  /* ------------------------------------------------------- */
  else
    {
      pid = psettings->pid;
      PROCMON_info_init( &pinfo );
      ierr_pget = PROCMON_info_get( pid, &pinfo );
      /* ----------- */
      /* valid entry */
      /* ----------- */
      if( ierr_pget != PROCMON_ERROR )
        {
          plist = calloc( 1, sizeof( PROCMON_INFO_LIST_struct ) );
          plist->pid = pid;
          plist->call_get_info = 0;
          PROCMON_info_copy( &pinfo, &(plist->pinfo) );
          if( strlen( psettings->outfile_prefix ) > 0 )
            {
              sprintf( plist->outfile_name, "%s%llu%s",
                       psettings->outfile_prefix, plist->pid, ".txt" );
              plist->outfile = fopen( plist->outfile_name, "a" );
            }
          else
            {
              plist->outfile = NULL;
            }
          *plist_head = plist;
        }
    }
  return( ierr );
}
/* ========================================================================

   Name
   ====
   procmon

   Purpose
   =======
   Standalone tool used somewhat like ps to get process info.

   Arguments
   =========
   argc          Intent: in
                 C type: int
                 Argument count.

   argv          Intent: in
                 C type: char**
                 Arguments

   ierr          Intent: out
                 C type: (na) error return value
                 error return value

   Return Values
   =============
   0 if ok

   Program Flow
   ============
   1) get settings
   2) set outfile
   3) print settings and header
   4) gather/print samples
   4.1) PROCMON_info_get
   4.2) if- PROCMON_info_print
   4.3) PROCMON_sleep

   ======================================================================== */
int main( int argc, const char **argv )
{
  /* --------- */
  /* variables */
  /* --------- */
  struct winsize ws;
  int ierr = 0; /* error return value */
  int ierr_pget; /* error for PROCMON_info_get */
  int lines_printed; /* number of lines printed (for pretty) */
  int printed_header; /* whether or not the header was printed */
  int width; /* width of screen */
  PROCMON_INFO_SETTINGS_struct psettings; /* settings */
  PROCMON_INFO_LIST_struct *plist_head; /* points to head of list */
  PROCMON_INFO_LIST_struct *plist_previous; /* points to previous in list */
  PROCMON_INFO_LIST_struct *plist_next; /* points to next in list */
  PROCMON_INFO_LIST_struct *plist; /* points to a member of list */
  /* --------------------- */
  /* executable statements */
  /* --------------------- */
  if( argc == 1 )
    {
      printf( "\n" );
      printf(
             "Usage: procmon [options]\n"
             "\n"
             "Options:\n"
             " If the option can also be set as an environment variable,\n"
             " the name will be in parenthesi at the end of the option.\n"
             "   [pid]\n"
             "      Process ID to sample.\n"
             "      [pid] overrides -uid and -command options.\n"
             "   -command <command> (PROCMON_COMMAND)\n"
             "      Gather info about all processes with the command.\n"
             "      Processes that start after this command is executed\n"
             "      will not be sampled.\n"
             "      Options -uid and -command will be intersected.\n"
             "      [pid] overrides -uid and -command options.\n"
             "   -all (PROCMON_ALL)\n"
             "      Get information about all processes.\n"
             "      Processes that start after this command is executed\n"
             "      will not be sampled.\n"
             "   -dir <output dir> (PROCMON_DIR)\n"
             "      Outpt directory for files.\n"
             "      File Format: <output dir>/procmon_<machine>_<pid>\n"
             "      Default: Print to stdout.\n"
             "      Specifying -dir turns off -pretty.\n"
             "   -interval <n> (PROCMON_INTERVAL)\n"
             "      Print every <n-th> sampling.\n"
             "      Default: (1) Print every sampling.\n"
             "      If set to 0, only print at process termination.\n"
             "   -min <n> (PROCMON_MIN)\n"
             "      Delete files unless got this min number of samples.\n"
             "      Default: (1) Keep every file with at least 1 sample.\n"
             "      This is most useful when using the Run Time Linker\n"
             "      Environment Variable approach.  A process might spawn\n"
             "      many other processes that you are not interested in.\n"
             "   -pretty (PROCMON_PRETTY)\n"
             "      Print pretty output - for interactive viewing\n"
             "      Default: Off\n"
             "      Make sure your screen is wide enough.\n"
             "      Specifying -dir turns off -pretty.\n"
             "   -prettyw <n> (PROCMON_PRETTYW)\n"
             "      Manually specify screen width.\n"
             "      Default: (use ioctl - if fail use 80).\n"
             "      Sometimes, the screen width cannot be obtained (implies\n"
             "      the -pretty option).  In this case, you can\n"
             "      specify the width manually.\n"
             "   -signal <int> (PROCMON_SIGNAL)\n"
             "      Signal to use for sampling.\n"
             "      Default: (SIGRTMAX if using posix itimer)\n"
             "                SIGALRM otherwise)\n"
             "      Currently, only used in automatic instrumentation.\n"
             "      An integer value must be used.\n"
             "   -time <millisecs> (PROCMON_TIME)\n"
             "      Time interval between samplings.\n"
             "      Default: (0) Sample and print once instead of sampling\n"
             "               until process ends.\n"
             "   -uid <n> (PROCMON_UID)\n"
             "      Gather info about all processes with the uid.\n"
             "      Processes that start after this command is executed\n"
             "      will not be sampled.\n"
             "      Options -uid and -command will be intersected.\n"
             "      [pid] overrides -uid and -command options.\n"
             "\n"
             "Post Processing:\n"
             "   When finished, procmon_post.pl can be used to generate\n"
             "   plots of the output file(s).  For example:\n"
             "      procmon_post.pl procmon_*.txt\n"
             "\n"
             );
      return( 0 );
    }
  /* ------------ */
  /* get settings */
  /* ------------ */
  ierr = PROCMON_info_settings_get( argc, argv, &psettings );
  assert( !ierr );
  /* ----------------- */
  /* create pinfo list */
  /* ----------------- */
  plist_head = NULL;
  PROCMON_list_create( &psettings, &plist_head );
  /* ----------------------------- */
  /* get width of screen if pretty */
  /* ----------------------------- */
  width = 0;
  if( psettings.pretty != 0 )
    {
      if( psettings.prettyw != 0 )
        {
          width = psettings.prettyw;
        }
      else if( ioctl( 1, TIOCGWINSZ, &ws ) >= 0 )
        {
          width = ws.ws_col;
        }
      else
        {
          /* just guess */
          width = 80;
        }
    }
  /* -------------- */
  /* print settings */
  /* -------------- */
  if( plist_head != NULL )
    {
      PROCMON_info_settings_print( &psettings, plist_head->outfile );
    }
  /* -------------------- */
  /* gather/print samples */
  /* -------------------- */
  printed_header = 0;
  lines_printed = 0;
  while( plist_head != NULL )
    {
      /* ---------------------------------------- */
      /* go up to overwrite lines if doing pretty */
      /* ---------------------------------------- */
      if( psettings.pretty != 0 && lines_printed > 0 )
        {
          fprintf( stdout, "\033[%dA", lines_printed );
        }
      lines_printed = 0;
      /* ------------ */
      /* process list */
      /* ------------ */
      plist = plist_head;
      plist_previous = NULL;
      while( plist )
        {
          /* --------------------------- */
          /* get new info or use old one */
          /* --------------------------- */
          if( plist->call_get_info )
            {
              ierr_pget = PROCMON_info_get( plist->pid, &(plist->pinfo) );
            }
          else
            {
              ierr_pget = plist->ierr_pget;
              plist->call_get_info = 1;
            }
          /* ------------------------------- */
          /* print if:                       */
          /* - interval and good data        */
          /* - time == 0 (only printing one) */
          /* - PROCMON_ERROR (process gone)  */
          /* ------------------------------- */
          if( ( ierr_pget == 0 &&
                psettings.interval > 0 &&
                plist->pinfo.num_samples % psettings.interval == 0 ) ||
              ( psettings.time == 0 ) ||
              ( ierr_pget == PROCMON_ERROR ) )
            {
              /* -------------------------------------- */
              /* only print header once if doing pretty */
              /* -------------------------------------- */
              if( psettings.pretty != 0 )
                {
                  plist->pinfo.printed_header = printed_header;
                }
              ierr = PROCMON_info_print( &(plist->pinfo), &psettings,
                                         (plist->outfile) );
              /* ---------------------------------------- */
              /* figure out lines printed if doing pretty */
              /* ---------------------------------------- */
              if( psettings.pretty != 0 )
                {
                  if( width > 0 )
                    {
                      lines_printed +=
                        ((plist->pinfo.chars_printed-1)/width) + 1;
                      if( (plist->pinfo.chars_printed-1) % width == 0 )
                        {
                          lines_printed--;
                        }
                    }
                  else
                    {
                      lines_printed++;
                    }
                }
              printed_header = 1;
            }
          /* ------------------------ */
          /* if this is last sampling */
          /* ------------------------ */
          if( psettings.time == 0 || ierr_pget == PROCMON_ERROR )
            {
              /* close file */
              if( plist->outfile != NULL )
                {
                  fclose( plist->outfile );
                  plist->outfile = NULL;
                  if( plist->pinfo.num_samples < psettings.min ||
                      plist->pinfo.num_samples_printed < 1 )
                    {
                      ierr = remove( plist->outfile_name );
                      assert( !ierr );
                    }
                }
              /* free from list */
              if( plist == plist_head )
                {
                  plist_head = plist->next;
                }
              else
                {
                  plist_previous->next = plist->next;
                }
              plist_next = plist->next;
              free( plist );
              plist = plist_next;
              continue;
            }
          plist_previous = plist;
          plist = plist->next;
        }
      /* ------------------ */
      /* DONE: process list */
      /* ------------------ */
      /* ------------------------- */
      /* sleep until next sampling */
      /* ------------------------- */
      ierr = PROCMON_sleep( psettings.time/1000.0 );
    }
  /* -------------------------- */
  /* DONE: gather/print samples */
  /* -------------------------- */
  return( ierr );
}
/* ========================================================================

   Name
   ====
   PROCMON_sleep

   Purpose
   =======
   fine-grained sleep timer

   Arguments
   =========
   time          Intent: in
                 C type: double
                 time in seconds to sleep.

   ierr          Intent: out
                 C type: (na) error return value
                 error return value

   Return Values
   =============
   0 if ok

   Program Flow
   ============
   1) sleep

   ======================================================================== */
int PROCMON_sleep( double time )
{
  int ierr = 0; /* error return value */
#if defined( _POSIX_TIMERS )
  struct timespec sleep_time_fine; /* sent to fine sleep timer */
#else
  useconds_t sleep_time_fine; /* sent to fine sleep timer */
#endif
  int sleep_time_secs; /* secs to sleep */
  int sleep_time_usecs; /* usecs to sleep */
  /* -------------------------------------- */
  /* creat secs and part and sleep for secs */
  /* -------------------------------------- */
  sleep_time_secs = (int)time;
  sleep_time_usecs = (time - sleep_time_secs) * 1000000.0;
  if( sleep_time_secs > 0 )
    {
      sleep( sleep_time_secs );
    }
  /* -------------- */
  /* sleep for part */
  /* -------------- */
#if defined( _POSIX_TIMERS )
  sleep_time_fine.tv_sec = 0;
  sleep_time_fine.tv_nsec = sleep_time_usecs * 1000;
  nanosleep( &sleep_time_fine, NULL );
#else
  usleep( sleep_time_usecs );
#endif
  return( ierr );
}
